from typing import List, Optional

from twitwi.bluesky.types import BlueskyProfile, BlueskyPartialProfile, BlueskyPost

PROFILE_FIELDS = list(BlueskyProfile.__annotations__.keys())

PARTIAL_PROFILE_FIELDS = list(BlueskyPartialProfile.__annotations__.keys())

POST_FIELDS = list(BlueskyPost.__annotations__.keys())

POST_PLURAL_FIELDS = [
    k
    for k, v in BlueskyPost.__annotations__.items()
    if v == List[str] or v == Optional[List[str]]
]

POST_BOOLEAN_FIELDS = [
    k
    for k, v in BlueskyPost.__annotations__.items()
    if v is bool or v == Optional[bool]
]



# =============================================================================
# Twitwi Constants
# =============================================================================
#
# Useful constants used throughout the library.
#
SOURCE_DATETIME_FORMAT = "%a %b %d %H:%M:%S +0000 %Y"
SOURCE_DATETIME_FORMAT_V2 = "%Y-%m-%dT%H:%M:%S.%fZ"
FORMATTED_TWEET_DATETIME_FORMAT = "%Y-%m-%dT%H:%M:%S"

FORMATTED_FULL_DATETIME_FORMAT = "%Y-%m-%dT%H:%M:%S.%f"

# More details on Twitter's tweets metadata can be read here: https://developer.twitter.com/en/docs/tweets/data-dictionary/overview/tweet-object
TWEET_FIELDS = [
    "id",                       # digital ID
    "timestamp_utc",            # UNIX timestamp of creation - UTC time
    "local_time",               # ISO datetime of creation - local time
    "user_screen_name",         # author's user text ID (@user) (at collection time)
    "text",                     # message's text content
    # 'filter_level',           # maximum value of the filter_level parameter which may be used and still stream this Tweet
    "possibly_sensitive",       # whether a link present in the message might contain sensitive content according to Twitter
    # 'withheld_copyright',     # whether the tweet might be censored by Twitter following copyright requests, ignorable
    # 'withheld_scope',         # whether the content withheld is the 'status' or a 'user', ignorable
    # 'withheld_countries',     # list of ISO country codes in which the message is withheld, separated by |, ignorable
    # 'truncated',              # whether the tweet is bigger than 140 characters, obsolete
    "retweet_count",            # number of retweets of the message (at collection time)
    "like_count",               # number of likes of the message (at collection time)
    "reply_count",              # number of answers to the message, dropped by Twitter (since Oct 17, now charged), unreliable and ignorable
    "impression_count",         # number of impressions generated by the message (at collection time)
    "lang",                     # language of the message automatically identified by Twitter's algorithms (equals 'und' when no language could be detected)
    "to_username",              # text ID of the user the message is answering to
    "to_userid",                # digital ID of the user the message is answering to
    "to_tweetid",               # digital ID of the tweet the message is answering to
    # 'source',                 # medium used by the user to post the message, now exported in source_name and source_url fields
    "source_name",              # name of the medium used to post the message
    "source_url",               # link to the medium used to post the message
    "user_location",            # location declared in the user's profile (at collection time)
    "lat",                      # latitude of messages geolocalized
    "lng",                      # longitude of messages geolocalized
    "user_id",                  # author's user digital ID
    "user_name",                # author's detailed textual name (at collection time)
    "user_verified",            # whether the author's account is certified
    "user_description",         # description given in the author's profile (at collection time)
    "user_url",                 # link to a website given in the author's profile (at collection time)
    "user_image",               # link to the image avatar of the author's profile (at collection time)
    # 'user_utcoffset',         # time offset due to the user's timezone, dropped by Twitter (since May 2018), ignorable
    # 'user_timezone',          # timezone declared in the user's profile, dropped by Twitter (since May 2018), ignorable
    # 'user_lang',              # language declared in the user's profile (at collection time), dropped by Twitter (since May 2019), ignorable
    "user_tweets",              # number of tweets sent by the user (at collection time)
    "user_followers",           # number of users following the author (at collection time)
    "user_friends",             # number of users the author is following (at collection time)
    "user_likes",               # number of likes the author has expressed (at collection time)
    "user_lists",               # number of users lists the author has been included in (at collection time)
    "user_created_at",          # ISO datetime of creation of the author's account
    "user_timestamp_utc",       # UNIX timestamp of creation of the author's account - UTC time
    "collected_via",            # How we received the message: 'stream', 'search', 'retweet' (the original tweet was
                                # contained in the retweet metadata), 'quote' (the original tweet was contained in
                                # the quote metadata), 'thread' (the tweet is part of the same conversation as a
                                # tweet collected via search or stream). If the message was collected via multiple
                                # ways, they are separated by |
    "match_query",              # whether the tweet was retrieved because it matches the query, or whether it was
                                # collected via 'quote' or 'thread'
    "retweeted_id",             # digital ID of the retweeted message
    "retweeted_user",           # text ID of the user who authored the retweeted message
    "retweeted_user_id",        # digital ID of the user who authoring the retweeted message
    "retweeted_timestamp_utc",  # UNIX timestamp of creation of the retweeted message - UTC time
    "quoted_id",                # digital ID of the retweeted message
    "quoted_user",              # text ID of the user who authored the quoted message
    "quoted_user_id",           # digital ID of the user who authoring the quoted message
    "quoted_timestamp_utc",     # UNIX timestamp of creation of the quoted message - UTC time
    "collection_time",          # ISO datetime of message collection - local time
    "url",                      # url of the tweet (to get a view of the message directly on Twitter)
    "place_country_code",       # if the tweet has an associated 'place', country code of that place
    "place_name",               # if the tweet has an associated 'place', name of that place
    "place_type",               # if the tweet has an associated 'place', type of that place ('city', 'admin', etc.)
    "place_coordinates",        # if the tweet has an associated 'place', coordinates of that place, separated by |
    "links",                    # list of links included in the text content, with redirections resolved, separated by |
    "domains",                  # list of domain names in the links fields, separated by |
    "media_urls",               # list of links to images/videos embedded, separated by |
    "media_files",              # list of filenames of images/videos embedded and downloaded, separated by |, ignorable when medias collections isn't enabled
    "media_types",              # list of media types (photo, video, animated gif), separated by |
    "media_alt_texts",          # list of alternative texts (image descriptions), separated by |
    "mentioned_names",          # list of text IDs of users mentionned, separated by |
    "mentioned_ids",            # list of digital IDs of users mentionned, separated by |
    "hashtags",                 # list of hashtags used, lowercased, separated by |
]

TWEET_FIELDS_TCAT = [
    "id",
    "time",
    "created_at",
    "from_user_name",
    "text",
    "filter_level",
    "possibly_sensitive",
    "withheld_copyright",
    "withheld_scope",
    "truncated",
    "retweet_count",
    "favorite_count",
    "lang",
    "to_user_name",
    "in_reply_to_status_id",
    "quoted_status_id",
    "source",
    "location",
    "lat",
    "lng",
    "from_user_id",
    "from_user_realname",
    "from_user_verified",
    "from_user_description",
    "from_user_url",
    "from_user_profile_image_url",
    "from_user_utcoffset",
    "from_user_timezone",
    "from_user_lang",
    "from_user_tweetcount",
    "from_user_followercount",
    "from_user_friendcount",
    "from_user_favourites_count",
    "from_user_listed",
    "from_user_withheld_scope",
    "from_user_created_at",
    "urls",
    "urls_expanded",
    "urls_followed",
    "domains",
    "HTTP status code",
    "media_id",
    "media_urls",
    "media_type",
    "media_indice_start",
    "media_indice_end",
    "photo_sizes_width",
    "photo_sizes_height",
    "photo_resize",
    "mentions",
    "hashtags",
]

GAZOU_TO_TCAT = {
    "identical_fields": {
        "id": "id",
        "timestamp_utc": "time",
        "local_time": "created_at",
        "user_screen_name": "from_user_name",
        "text": "text",
        "possibly_sensitive": "possibly_sensitive",
        "retweet_count": "retweet_count",
        "like_count": "favorite_count",
        "lang": "lang",
        "to_username": "to_user_name",
        "to_userid": "to_user_id",
        "to_tweetid": "in_reply_to_status_id",
        "quoted_id": "quoted_status_id",
        "user_location": "location",
        "lat": "lat",
        "lng": "lng",
        "user_id": "from_user_id",
        "user_name": "from_user_realname",
        "user_verified": "from_user_verified",
        "user_description": "from_user_description",
        "user_url": "from_user_url",
        "user_image": "from_user_profile_image_url",
        "user_tweets": "from_user_tweetcount",
        "user_followers": "from_user_followercount",
        "user_friends": "from_user_friendcount",
        "user_likes": "from_user_favourites_count",
        "user_lists": "from_user_listed",
        "user_created_at": "from_user_created_at",
        "links": "urls_expanded",
        "domains": "domains",
        "mentioned_ids": "mentions",
        "hashtags": "hashtags",
        "media_urls": "media_urls",
        "media_types": "media_type",
    },
    "modified_fields": ["source"],
    "removed_fields": [
        "filter_level",
        "witheld_copyright",
        "withheld_scope",
        "truncated",
        "from_user_utcoffset",
        "from_user_timezone",
        "from_user_lang",
        "from_user_withheld_scope",
        "urls",
        "media_id",
        "media_indice_start",
        "media_indice_end",
        "photo_sizes_width",
        "photo_sizes_height",
        "photo_resize",
    ],
}

TWEET_PLURAL_FIELDS = {
    "links",
    "urls_expanded",
    "domains",
    "hashtags",
    "collected_via",
    "media_urls",
    "media_files",
    "media_types",
    "media_alt_texts",
    "mentioned_names",
    "mentioned_ids",
    "mentions",
}

TWEET_BOOLEAN_FIELDS = {"possibly_sensitive", "user_verified", "match_query"}

# More details on Twitter's users metadata can be read here: https://developer.twitter.com/en/docs/tweets/data-dictionary/overview/user-object
USER_FIELDS = [
    "id",
    "screen_name",
    "name",
    "description",
    "url",
    # 'lang',                               # dropped from tweet objects only by Twitter (since May 2019)
    "timestamp_utc",
    "local_time",
    # 'utc_offset',                         # dropped by Twitter (since May 2018), ignorable
    # 'time_zone',                          # dropped by Twitter (since May 2018), ignorable
    "location",
    # 'geo_enabled',                        # dropped by Twitter (since May 2019), ignorable
    "verified",
    "protected",
    "tweets",
    "followers",
    "friends",
    "likes",
    "lists",
    # 'is_translator',                      # dropped by Twitter (since May 2019), ignorable
    # 'translator_type',                    # dropped by Twitter (since May 2019), ignorable
    # 'is_translation_enabled',             # dropped by Twitter (since May 2019), ignorable
    # 'default_profile',
    # 'default_profile_image',
    # 'has_extended_profile',               # dropped by Twitter (since May 2019), ignorable
    # 'profile_image_url',                  # dropped by Twitter (since May 2019), ignorable
    "image",
    # 'profile_banner_url',
    # 'profile_use_background_image',       # dropped by Twitter (since May 2019), ignorable
    # 'profile_background_image_url',       # dropped by Twitter (since May 2019), ignorable
    # 'profile_background_image_url_https', # dropped by Twitter (since May 2019), ignorable
    # 'profile_background_tile',            # dropped by Twitter (since May 2019), ignorable
    # 'profile_background_color',           # dropped by Twitter (since May 2019), ignorable
    # 'profile_link_color',                 # dropped by Twitter (since May 2019), ignorable
    # 'profile_text_color',                 # dropped by Twitter (since May 2019), ignorable
    # 'profile_sidebar_fill_color',         # dropped by Twitter (since May 2019), ignorable
    # 'profile_sidebar_border_color'        # dropped by Twitter (since May 2019), ignorable
    "default_profile",
    "default_profile_image",
    "witheld_in_countries",
    "witheld_scope",
]

USER_PLURAL_FIELDS = {"witheld_in_countries"}

USER_BOOLEAN_FIELDS = {
    "verified",
    "protected",
    "default_profile",
    "default_profile_image",
}

CANONICAL_URL_KWARGS = {
    "strip_authentication": False,
    "strip_trailing_slash": False,
    "strip_protocol": False,
    "strip_irrelevant_subdomains": False,
    "strip_fragment": False,
    "normalize_amp": False,
    "fix_common_mistakes": False,
    "infer_redirection": False,
    "quoted": True,
}

CANONICAL_HOSTNAME_KWARGS = {"normalize_amp": False, "infer_redirection": False}

# API v2 constants
TWEET_FIELDS_V2 = {
    "attachments",
    "author_id",
    # NOTE: (2023-04-26) dropping this because we don't use it and it prevents us
    # from being able to get 500 tweets per call using academic v2 API.
    # 'context_annotations',
    "conversation_id",
    "created_at",
    "entities",
    "geo",
    "id",
    "in_reply_to_user_id",
    "lang",
    "possibly_sensitive",
    "public_metrics",
    "referenced_tweets",
    "reply_settings",
    "source",
    "text",
    "withheld",
}

MEDIA_FIELDS = {
    "media_key",
    "type",
    "duration_ms",
    "height",
    "preview_image_url",
    "public_metrics",
    "width",
    "alt_text",
    "url",
    "variants",
}

POLL_FIELDS = {"id", "options", "duration_minutes", "end_datetime", "voting_status"}

PLACE_FIELDS = {
    "full_name",
    "id",
    "contained_within",
    "country",
    "country_code",
    "geo",
    "name",
    "place_type",
}

USER_FIELDS_V2 = {
    "id",
    "name",
    "username",
    "created_at",
    "description",
    "entities",
    "location",
    "pinned_tweet_id",
    "profile_image_url",
    "protected",
    "public_metrics",
    "url",
    "verified",
    "withheld",
}

TWEET_EXPANSIONS = {
    "author_id",
    "referenced_tweets.id",
    "in_reply_to_user_id",
    "attachments.media_keys",
    "attachments.poll_ids",
    "geo.place_id",
    "entities.mentions.username",
    "referenced_tweets.id.author_id",
}

TWEET_PARAMS = {
    "tweet.fields": ",".join(TWEET_FIELDS_V2),
    "media.fields": ",".join(MEDIA_FIELDS),
    "poll.fields": ",".join(POLL_FIELDS),
    "place.fields": ",".join(PLACE_FIELDS),
    "user.fields": ",".join(USER_FIELDS_V2),
}

USER_EXPANSIONS = {"pinned_tweet_id"}

USER_PARAMS = {
    "user.fields": ",".join(USER_FIELDS_V2),
    "tweet.fields": ",".join(TWEET_FIELDS_V2),
}

# Lists

LIST_FIELDS = {
    "created_at",
    "follower_count",
    "member_count",
    "private",
    "description",
    "owner_id",
}

LIST_EXPANSIONS = {"owner_id"}

LIST_PARAMS = {
    "list.fields": ",".join(LIST_FIELDS),
    "user.fields": ",".join(USER_FIELDS_V2),
}

LIST_TWEETS_EXPANSIONS = {"author_id"}

LIST_MEMBERS_EXPANSIONS = {"pinned_tweet_id"}

LIST_TWEETS_OR_MEMBERS_PARAMS = {
    "tweet.fields": ",".join(TWEET_FIELDS_V2),
    "user.fields": ",".join(USER_FIELDS_V2),
}

PRE_SNOWFLAKE_LAST_TWEET_ID = 29700859247
OFFSET_TIMESTAMP = 1288834974657



import re

from twitwi.exceptions import BlueskyPayloadError


valid_post_keys = [
    "cid",
    "uri",
    "author",
    "record",
    "replyCount",
    "repostCount",
    "likeCount",
    "quoteCount",
]

valid_record_keys = ["$type", "createdAt", "text"]


valid_author_keys = ["did", "handle", "createdAt"]


def validate_post_payload(data):
    post = data.get("post", data)

    for key in valid_post_keys:
        if key not in post:
            return False, f"key {key} is missing from payload: {post}"

    if not isinstance(post["record"], dict):
        return False, "payload's record is not a dictionary: %s" % post["record"]

    for key in valid_record_keys:
        if key not in post["record"]:
            return False, "key %s is missing from payload's record: %s" % (
                key,
                post["record"],
            )

    # Splitting by '#' to ignore possible suffixes in $type
    # e.g. https://bsky.app/profile/did:plc:k6acu4chiwkixvdedcmdgmal/post/3lagdncjsu22y
    if post["record"].get("$type").split("#")[0] != "app.bsky.feed.post":
        return False, "payload's record $type is not a post: %s" % post["record"].get(
            "$type"
        )

    if not isinstance(post["author"], dict):
        return False, "payload's author is not a dictionary: %s" % post["author"]

    for key in valid_author_keys:
        if key not in post["author"]:
            return False, "key %s is missing from payload's author: %s" % (
                key,
                post["author"],
            )

    return True, None


re_embed_types = re.compile(
    r"(?:\.(?:record|recordWithMedia|images|videos?|external|post|embed|links|media|file|viewImages)(?:#.*)?|N\/A|image)$"
)


def valid_embed_type(embed_type):
    return re_embed_types.search(embed_type)


def format_profile_url(user_handle_or_did):
    return f"https://bsky.app/profile/{user_handle_or_did}"


def format_post_url(user_handle_or_did, post_did, post_splitter="/post/"):
    return f"https://bsky.app/profile/{user_handle_or_did}{post_splitter}{post_did}"


def parse_post_url(url, source):
    """Returns a tuple of (author_handle/did, post_did) from an https://bsky.app post URL"""

    known_splits = ["/post/", "/lists/", "/feed/"]

    if url.startswith("https://bsky.app/profile/"):
        for split in known_splits:
            if split in url[25:]:
                return url[25:].split(split)

    raise BlueskyPayloadError(source, f"{url} is not a usual Bluesky post url")


def parse_post_uri(uri, source=None):
    """Returns a tuple of (author_did, post_did) from an at:// post URI"""

    # known_splits = [
    #     "/app.bsky.feed.post/",
    #     "/app.bsky.graph.starterpack/",
    #     "/app.bsky.feed.generator/",
    #     "/app.bsky.graph.list/",
    #     "/app.bsky.graph.follow/", # This one is often found when a post is an anwser to a deleted post (e.g. https://bsky.app/profile/sydney-chat.bsky.social/post/3ltsph6kxfl25)
    # ]

    # if uri.startswith("at://"):
    #     for split in known_splits:
    #         if split in uri:
    #             return uri[5:].split(split)

    # There's too much variability in the post URIs, and we cannot be exhaustive,
    # so we do with the simple approach:
    if uri.startswith("at://"):
        # Using maxsplit=3 to avoid issues if future uris contain more slashes
        author_did, _, post_did = uri[5:].split("/", 3)
        return author_did, post_did

    raise BlueskyPayloadError(source or uri, f"{uri} is not a usual Bluesky post uri")


def format_starterpack_url(user_handle_or_did, record_did):
    return f"https://bsky.app/starter-pack/{user_handle_or_did}/{record_did}"


def format_media_url(user_did, media_cid, mime_type, source):
    media_type = mime_type.split("/")[1]
    if mime_type.startswith("image"):
        media_url = f"https://cdn.bsky.app/img/feed_fullsize/plain/{user_did}/{media_cid}@{media_type}"
        media_thumb = f"https://cdn.bsky.app/img/feed_thumbnail/plain/{user_did}/{media_cid}@{media_type}"
    elif (
        mime_type.startswith("video")
        or mime_type == "application/xml"
        or mime_type == "*/*"
    ):
        media_url = f"https://video.bsky.app/watch/{user_did}/{media_cid}/playlist.m3u8"
        media_thumb = (
            f"https://video.bsky.app/watch/{user_did}/{media_cid}/thumbnail.jpg"
        )
    elif any(mt in mime_type for mt in ["octet-stream", "text/plain", "text/html"]):
        media_url = (
            f"https://cdn.bsky.app/img/feed_fullsize/plain/{user_did}/{media_cid}@jpeg"
        )
        media_thumb = (
            f"https://cdn.bsky.app/img/feed_thumbnail/plain/{user_did}/{media_cid}@jpeg"
        )
    elif "empty" in mime_type:
        media_url, media_thumb = "", ""
    else:
        raise BlueskyPayloadError(source, f"{mime_type} is an unusual media mimeType")
    return media_url, media_thumb



# =============================================================================
# Twitwi Library Endpoint
# =============================================================================
#
from twitwi.anonymizers import anonymize_normalized_tweet
from twitwi.formatters import (
    transform_tweet_into_csv_dict,
    format_tweet_as_csv_row,
    transform_user_into_csv_dict,
    format_user_as_csv_row,
    apply_tcat_format,
)

# NOTE: should we drop this from public exports?
from twitwi.utils import (
    get_dates,
    custom_normalize_url,
    get_timestamp_from_id,
    get_dates_from_id,
)
from twitwi.normalizers import (
    normalize_tweet,
    normalize_user,
    normalize_tweets_payload_v2,
)

__all__ = [
    "anonymize_normalized_tweet",
    "transform_tweet_into_csv_dict",
    "format_tweet_as_csv_row",
    "transform_user_into_csv_dict",
    "format_user_as_csv_row",
    "apply_tcat_format",
    "get_dates",
    "custom_normalize_url",
    "get_timestamp_from_id",
    "get_dates_from_id",
    "normalize_tweet",
    "normalize_user",
    "normalize_tweets_payload_v2",
]



# =============================================================================
# Twitwi Exceptions
# =============================================================================
#
# Custom exceptions used by the library.
#


class TwitwiError(Exception):
    pass


class TwitterPayloadV2IncompleteIncludesError(TwitwiError):
    def __init__(self, kind, key):
        self.kind = kind
        self.key = key
        super().__init__("{!r} ({})".format(key, kind))


class BlueskyPayloadError(TwitwiError):
    def __init__(self, source, message):
        self.source = source
        self.message = message
        super().__init__(f"Error while processing Bluesky post {source}.\n{message}")



# =============================================================================
# Twitwi Formatters
# =============================================================================
#
# Miscellaneous formatter functions transforming tweets from one format
# to the other.
#
from twitwi.constants import (
    TWEET_FIELDS,
    GAZOU_TO_TCAT,
    TWEET_BOOLEAN_FIELDS,
    TWEET_PLURAL_FIELDS,
    USER_FIELDS,
    USER_BOOLEAN_FIELDS,
    USER_PLURAL_FIELDS,
)


def apply_tcat_format(item):
    result = {
        v: item[k] for k, v in GAZOU_TO_TCAT["identical_fields"].items() if k in item
    }
    result["source"] = "<a href={} rel=nofollow>{}</a>".format(
        item["source_url"], item["source_name"]
    )
    return result


def make_transform_into_csv_dict(plural_fields, boolean_fields):
    def transform_into_csv_dict(
        item, item_id=None, plural_separator="|", allow_erroneous_plurals=False
    ):
        if item_id is not None:
            item["id"] = item_id

        item["links"] = item.get("proper_links", item.get("links", []))

        for plural_field in plural_fields:
            plurals = item.get(plural_field, [])
            if allow_erroneous_plurals:
                plurals = [
                    element if element is not None else "" for element in plurals
                ]
            item[plural_field] = plural_separator.join(plurals)

        for boolean_field in boolean_fields:
            item[boolean_field] = (
                int(item[boolean_field]) if boolean_field in item else ""
            )

    return transform_into_csv_dict


def make_format_as_csv_row(fields, plural_fields, boolean_fields):
    def format_field_for_csv(
        field, item, item_id=None, plural_separator="|", allow_erroneous_plurals=False
    ):
        if field == "id" and item_id is not None:
            return item_id

        # NOTE: this is hardly the most performant way to proceed by I am lazy...
        if field in plural_fields:
            v = item.get(field, [])

            if field == "links":
                v = item.get("proper_links", v)

            # Clean None values that may have slipped in, such as in the 'domains' field when
            # normalizing this Bluesky post: https://bsky.app/profile/did:plc:cs5qjcmnntogoahrrsagmg2z/post/3lvqhn7raq62v
            if allow_erroneous_plurals:
                v = [element if element is not None else "" for element in v]

            return plural_separator.join(v)

        if field in boolean_fields:
            return int(item[field]) if field in item else ""

        return item.get(field, "")

    def format_item_as_csv_row(
        item, item_id=None, plural_separator="|", allow_erroneous_plurals=False
    ):
        return [
            format_field_for_csv(
                field,
                item,
                item_id=item_id,
                plural_separator=plural_separator,
                allow_erroneous_plurals=allow_erroneous_plurals,
            )
            for field in fields
        ]

    return format_item_as_csv_row


transform_tweet_into_csv_dict = make_transform_into_csv_dict(
    TWEET_PLURAL_FIELDS, TWEET_BOOLEAN_FIELDS
)

format_tweet_as_csv_row = make_format_as_csv_row(
    TWEET_FIELDS, TWEET_PLURAL_FIELDS, TWEET_BOOLEAN_FIELDS
)

transform_user_into_csv_dict = make_transform_into_csv_dict(
    USER_PLURAL_FIELDS, USER_BOOLEAN_FIELDS
)

format_user_as_csv_row = make_format_as_csv_row(
    USER_FIELDS, USER_PLURAL_FIELDS, USER_BOOLEAN_FIELDS
)

__all__ = [
    "transform_tweet_into_csv_dict",
    "format_tweet_as_csv_row",
    "transform_user_into_csv_dict",
    "format_user_as_csv_row",
    "apply_tcat_format",
]



# =============================================================================
# Twitwi Bluesky Normalizers Unit Tests
# =============================================================================
from functools import partial
from pytz import timezone
from copy import deepcopy

from twitwi.bluesky import normalize_profile, normalize_partial_profile, normalize_post

from test.utils import get_json_resource


# Set to True to regenerate test results
OVERWRITE_TESTS = False


FAKE_COLLECTION_TIME = "2025-01-01T00:00:00.000000"


def set_fake_collection_time(dico):
    if "collection_time" in dico:
        dico["collection_time"] = FAKE_COLLECTION_TIME
    return dico


def compare_dicts(_id, d1, d2, ignore_fields=[]):
    for k in d2.keys():
        if k not in ignore_fields + ["collection_time"]:
            assert d1[k] == d2[k], (
                'Different value for key "%s" with payload data for "%s"'
                % (
                    k,
                    _id,
                )
            )

    for k in d1.keys():
        if k not in ignore_fields:
            assert k in d2, 'Missing key "%s" with payload data for "%s"' % (k, _id)


class TestNormalizers:
    def test_normalize_profile(self):
        tz = timezone("Europe/Paris")

        profiles = get_json_resource("bluesky-profiles.json")
        fn = partial(normalize_profile, locale=tz)

        if OVERWRITE_TESTS:
            from test.utils import dump_json_resource

            normalized_profiles = [
                set_fake_collection_time(fn(profile)) for profile in profiles
            ]
            dump_json_resource(normalized_profiles, "bluesky-normalized-profiles.json")

        expected = get_json_resource("bluesky-normalized-profiles.json")

        for idx, profile in enumerate(profiles):
            result = fn(profile)
            assert isinstance(result, dict)
            assert "collection_time" in result and isinstance(
                result["collection_time"], str
            )

            compare_dicts(profile["handle"], result, expected[idx])

    def test_normalize_profile_should_not_mutate(self):
        profile = get_json_resource("bluesky-profiles.json")[0]

        original_arg = deepcopy(profile)

        normalize_profile(profile)

        assert profile == original_arg

    def test_normalize_partial_profile(self):
        tz = timezone("Europe/Paris")

        profiles = get_json_resource("bluesky-partial-profiles.json")
        fn = partial(normalize_partial_profile, locale=tz)

        if OVERWRITE_TESTS:
            from test.utils import dump_json_resource

            normalized_profiles = [
                set_fake_collection_time(fn(profile)) for profile in profiles
            ]
            dump_json_resource(
                normalized_profiles, "bluesky-normalized-partial-profiles.json"
            )

        expected = get_json_resource("bluesky-normalized-partial-profiles.json")

        for idx, profile in enumerate(profiles):
            result = fn(profile)
            assert isinstance(result, dict)
            assert "collection_time" in result and isinstance(
                result["collection_time"], str
            )

            compare_dicts(profile["handle"], result, expected[idx])

    def test_normalize_partial_profile_should_not_mutate(self):
        profile = get_json_resource("bluesky-partial-profiles.json")[0]

        original_arg = deepcopy(profile)

        normalize_partial_profile(profile)

        assert profile == original_arg

    def test_normalize_post(self):
        tz = timezone("Europe/Paris")

        posts = get_json_resource("bluesky-posts.json")
        fn = partial(normalize_post, locale=tz)

        if OVERWRITE_TESTS:
            from test.utils import dump_json_resource

            normalized_posts = [
                [
                    set_fake_collection_time(p)
                    for p in fn(post, extract_referenced_posts=True)
                ]
                for post in posts
            ]
            dump_json_resource(normalized_posts, "bluesky-normalized-posts.json")

        expected = get_json_resource("bluesky-normalized-posts.json")

        # With referenced tweets
        for idx, post in enumerate(posts):
            result = fn(post, extract_referenced_posts=True)
            assert isinstance(result, list)
            assert set(p["uri"] for p in result) == set(p["uri"] for p in expected[idx])
            for idx2, p in enumerate(result):
                assert "collection_time" in p and isinstance(p["collection_time"], str)

                if "post" in post:
                    uri = post["post"]["uri"]
                else:
                    uri = post["uri"]
                compare_dicts(uri, p, expected[idx][idx2])

        # With single output
        for idx, post in enumerate(posts):
            result = fn(post)

            assert isinstance(result, dict)

            _id = p["uri"]
            compare_dicts(_id, result, expected[idx][-1])

        # With custom collection_source
        for post in posts:
            result = fn(post, collection_source="unit_test")

            assert result["collected_via"] == ["unit_test"]

    def test_normalize_post_should_not_mutate(self):
        post = get_json_resource("bluesky-posts.json")[0]

        original_arg = deepcopy(post)

        normalize_post(post)

        assert post == original_arg

    def test_normalize_post_should_be_normalized_across_sources(self):
        # handle same post from different sources (search, get_post and user_feed)
        pass

    def test_badly_formatted_posts_payload(self):
        pass
